
from __future__ import annotations
import numpy as np

def octahedral_adjacency() -> np.ndarray:
    A = np.zeros((6,6), dtype=int)
    opp = {0:1, 1:0, 2:3, 3:2, 4:5, 5:4}
    for i in range(6):
        for j in range(6):
            if i != j and j != opp[i]:
                A[i,j] = 1
    return A

def embed_octonion_axes_to_octahedron() -> np.ndarray:
    dirs = np.array([
        [1,0,0],
        [0,1,0],
        [0,0,1],
        [1,1,0],
        [1,0,1],
        [0,1,1],
        [1,1,1]
    ], dtype=float)
    dirs /= np.linalg.norm(dirs, axis=1, keepdims=True)
    oct_dirs = np.array([
        [1,0,0],[-1,0,0],
        [0,1,0],[0,-1,0],
        [0,0,1],[0,0,-1]
    ], dtype=float)
    M = np.zeros((7,6))
    for i in range(7):
        cosines = dirs[i] @ oct_dirs.T
        M[i] = np.abs(cosines)
    return M




from a003558.coupling import coupled_step

def test_coupling_is_near_unit():
    for k1 in range(1,8):
        for k2 in range(1,8):
            err = coupled_step(k1, k2)
            assert err < 1e-8




from __future__ import annotations
import numpy as np
from .octonions import octonion_mul, norm

def dyadic_clock_to_octonion(k: int, phase: float = 0.0) -> np.ndarray:
    theta = (np.pi / (2**k)) + phase
    v = np.zeros(8)
    v[0] = np.cos(theta)
    v[1] = np.sin(theta)
    return v

def coupled_step(k_brain: int, k_cosmos: int) -> float:
    a = dyadic_clock_to_octonion(k_brain, phase=0.0)
    b = dyadic_clock_to_octonion(k_cosmos, phase=np.pi/8)
    c = octonion_mul(a, b)
    return abs(norm(c) - norm(a)*norm(b))



import os
import pathlib

import matplotlib
matplotlib.use("Agg")  # headless backend voor CI/pytest
import matplotlib.pyplot as plt

from a003558.viz import plot_basis

def test_plot_basis_smoke(tmp_path: pathlib.Path):
    """
    Smoke-test: de plot moet zonder exceptions renderen en een bestand kunnen wegschrijven.
    """
    out_png = tmp_path / "basis.png"
    fig = plot_basis(save_path=str(out_png), show=False, title="Test Plot")
    assert fig is not None
    plt.close(fig)

    assert out_png.exists()
    assert out_png.stat().st_size > 0



import matplotlib.pyplot as plt
import numpy as np


def plot_basis(save_path: str | None = None, show: bool = True, title: str | None = None):
    """Plot de standaardbasis in het vlak."""
    fig, ax = plt.subplots()
    ax.quiver([0, 0], [0, 0], [1, 0], [0, 1], angles="xy", scale_units="xy", scale=1, color=["r", "g"])
    ax.set_xlim(-0.5, 1.5)
    ax.set_ylim(-0.5, 1.5)
    ax.set_aspect("equal")
    if title:
        ax.set_title(title)
    if save_path:
        fig.savefig(save_path, dpi=150)
    if show:
        plt.show()
    return fig


def plot_cycle(points, save_path: str | None = None, show: bool = True, title: str | None = None):
    """Plot een gesloten cyclus van punten in het vlak."""
    pts = np.array(points)
    fig, ax = plt.subplots()
    ax.plot(pts[:, 0], pts[:, 1], "o-", lw=1.5)
    ax.set_aspect("equal")
    if title:
        ax.set_title(title)
    if save_path:
        fig.savefig(save_path, dpi=150)
    if show:
        plt.show()
    return fig



import pathlib
import matplotlib
matplotlib.use("Agg")

from a003558.viz_octa import plot_octahedron_and_cube

def test_plot_octahedron_and_cube(tmp_path: pathlib.Path):
    out = tmp_path / "octa_cube.png"
    fig = plot_octahedron_and_cube(save_path=str(out), show=False, title="Octa-Cube")
    assert fig is not None
    assert out.exists() and out.stat().st_size > 0



from __future__ import annotations
import numpy as np

"""
Octonions via Cayley–Dickson from quaternions.

We represent an octonion x ∈ R^8 as a pair of quaternions (a, b),
with a = (x0, x1, x2, x3), b = (x4, x5, x6, x7).

Product (a,b) * (c,d) = ( a*c - conj(d)*b,  d*a + b*conj(c) )
Conjugate: conj(a,b) = (conj(a), -b)
Norm: ||(a,b)||^2 = ||a||^2 + ||b||^2

This guarantees:
- Alternativity: (xx)y = x(xy)
- Composition:   ||xy|| = ||x|| * ||y||
"""

# ---------- Quaternion helpers ----------

def _q_conj(q: np.ndarray) -> np.ndarray:
    q = np.asarray(q, dtype=float).reshape(4)
    out = q.copy()
    out[1:] *= -1.0
    return out

def _q_mul(p: np.ndarray, q: np.ndarray) -> np.ndarray:
    """
    Hamilton quaternions (1, i, j, k) with i^2 = j^2 = k^2 = ijk = -1.
    """
    p = np.asarray(p, dtype=float).reshape(4)
    q = np.asarray(q, dtype=float).reshape(4)
    p0, p1, p2, p3 = p
    q0, q1, q2, q3 = q
    return np.array([
        p0*q0 - p1*q1 - p2*q2 - p3*q3,
        p0*q1 + p1*q0 + p2*q3 - p3*q2,
        p0*q2 - p1*q3 + p2*q0 + p3*q1,
        p0*q3 + p1*q2 - p2*q1 + p3*q0,
    ], dtype=float)

def _q_norm2(q: np.ndarray) -> float:
    q = np.asarray(q, dtype=float).reshape(4)
    return float(np.dot(q, q))

# ---------- Octonions via Cayley–Dickson ----------

def octonion_mul(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Cayley–Dickson product in R^8.
    x, y: arrays of length 8, basis (1, e1..e7) reals.

    Returns x*y as an array of length 8.
    """
    x = np.asarray(x, dtype=float).reshape(8)
    y = np.asarray(y, dtype=float).reshape(8)

    a = x[:4]  # quaternion
    b = x[4:]  # quaternion
    c = y[:4]
    d = y[4:]

    ac = _q_mul(a, c)
    d_conj = _q_conj(d)
    left = ac - _q_mul(d_conj, b)

    da = _q_mul(d, a)
    c_conj = _q_conj(c)
    right = da + _q_mul(b, c_conj)

    out = np.empty(8, dtype=float)
    out[:4] = left
    out[4:] = right
    return out

def conj(x: np.ndarray) -> np.ndarray:
    """
    Octonion conjugate: \bar{(a,b)} = (conj(a), -b)
    """
    x = np.asarray(x, dtype=float).reshape(8)
    a = x[:4]
    b = x[4:]
    out = np.empty(8, dtype=float)
    out[:4] = _q_conj(a)
    out[4:] = -b
    return out

def norm(x: np.ndarray) -> float:
    """
    Euclidean norm ||x||. For octonions built by Cayley–Dickson:
      ||xy|| = ||x|| * ||y||
    """
    x = np.asarray(x, dtype=float).reshape(8)
    # Euclidean norm coincides with the composition norm here
    return float(np.linalg.norm(x))

def random_unit(seed: int | None = None) -> np.ndarray:
    """
    Random unit octonion (uniform on the 7-sphere).
    """
    rng = np.random.default_rng(seed)
    v = rng.normal(size=8)
    v /= np.linalg.norm(v)
    return v

# ---------- Optional convenience ----------

def zero() -> np.ndarray:
    return np.zeros(8, dtype=float)

def one() -> np.ndarray:
    z = np.zeros(8, dtype=float)
    z[0] = 1.0
    return z

def from_scalar_vector(s: float, v: np.ndarray) -> np.ndarray:
    v = np.asarray(v, dtype=float).reshape(7)
    out = np.empty(8, dtype=float)
    out[0] = float(s)
    out[1:] = v
    return out

def to_scalar_vector(x: np.ndarray) -> tuple[float, np.ndarray]:
    x = np.asarray(x, dtype=float).reshape(8)
    return float(x[0]), x[1:].copy()




import numpy as np
from a003558.octonions import octonion_mul, random_unit, norm

def test_alternativity_numeric():
    for _ in range(10):
        x = random_unit(seed=None)
        y = random_unit(seed=None)
        xx = octonion_mul(x,x)
        left = octonion_mul(xx, y)
        right = octonion_mul(x, octonion_mul(x, y))
        assert np.allclose(left, right, atol=1e-8)

def test_norm_multiplicative_numeric():
    for _ in range(10):
        x = random_unit(None); y = random_unit(None)
        xy = octonion_mul(x,y)
        assert np.isclose(norm(xy), norm(x)*norm(y), atol=1e-8)



from __future__ import annotations
import math

def staircase_level(n: int) -> int:
    """
    Kritische 'trapfunctie' op dyadische schalen.
    1 -> 1
    2..3 -> 2
    4..7 -> 3
    8..15 -> 4
    etc.
    """
    if n < 1:
        raise ValueError("n moet positief zijn")
    return 1 + (n.bit_length() - 1)


def horizon_bound(n: int, L: int) -> bool:
    """
    Horizon condition: if L = ord_{2n-1}(2) then n <= 2^{L-1},
    equivalently: L >= 1 + ceil(log2 n).
    """
    return L >= 1 + math.ceil(math.log2(n))



"""
Export routines to generate simple OBJ geometry for Blender or other 3D tools.
"""

from __future__ import annotations

import math
import os
from typing import Iterable


def _write_obj(path: str, vertices: list[tuple[float, float, float]], faces: list[tuple[int, int, int]]) -> str:
    """Write a minimal Wavefront OBJ file."""
    with open(path, "w", encoding="utf-8") as f:
        for v in vertices:
            f.write(f"v {v[0]} {v[1]} {v[2]}\n")
        for face in faces:
            f.write("f " + " ".join(str(i) for i in face) + "\n")
    return path


def export_octa_cube_obj(out_dir: str, basename: str = "scene") -> dict[str, str]:
    """
    Exporteer zowel een octaëder als een kubus naar OBJ-bestanden.
    Retourneert een dict met paden.

    Parameters
    ----------
    out_dir : str
        Doelmap.
    basename : str
        Basisnaam van de bestanden (zonder extensie).
    """
    os.makedirs(out_dir, exist_ok=True)

    # Eenhedenkubus (vertices van -1 tot 1)
    cube_vertices = [
        (-1, -1, -1),
        (-1, -1, 1),
        (-1, 1, -1),
        (-1, 1, 1),
        (1, -1, -1),
        (1, -1, 1),
        (1, 1, -1),
        (1, 1, 1),
    ]
    cube_faces = [
        (1, 2, 4),
        (1, 4, 3),
        (5, 7, 8),
        (5, 8, 6),
        (1, 5, 6),
        (1, 6, 2),
        (2, 6, 8),
        (2, 8, 4),
        (3, 4, 8),
        (3, 8, 7),
        (1, 3, 7),
        (1, 7, 5),
    ]

    octa_vertices = [
        (1, 0, 0),
        (-1, 0, 0),
        (0, 1, 0),
        (0, -1, 0),
        (0, 0, 1),
        (0, 0, -1),
    ]
    octa_faces = [
        (1, 3, 5),
        (3, 2, 5),
        (2, 4, 5),
        (4, 1, 5),
        (3, 1, 6),
        (2, 3, 6),
        (4, 2, 6),
        (1, 4, 6),
    ]

    cube_path = os.path.join(out_dir, f"{basename}_cube.obj")
    octa_path = os.path.join(out_dir, f"{basename}_octa.obj")

    _write_obj(cube_path, cube_vertices, cube_faces)
    _write_obj(octa_path, octa_vertices, octa_faces)

    return {"cube": cube_path, "octa": octa_path}


def export_label_spheres_obj(
    out_dir: str | None = None,
    basename: str = "labels",
    labels: Iterable[str] | None = None,
    path: str | None = None,
) -> str:
    """
    Exporteer labels als kleine sferen (hier gesimuleerd als punten) naar OBJ-bestand.

    Parameters
    ----------
    out_dir : str, optional
        Doelmap waarin geschreven wordt (indien `path` niet is opgegeven).
    basename : str
        Basisnaam van het bestand.
    labels : Iterable[str], optional
        Namen van labels (momenteel niet gebruikt in geometrie).
    path : str, optional
        Volledig pad naar output-bestand. Heeft voorrang op `out_dir` + `basename`.

    Returns
    -------
    str
        Pad naar geschreven bestand.
    """
    if path is not None:
        out_path = str(path)
    else:
        if out_dir is None:
            raise ValueError("Either `path` or `out_dir` must be provided")
        os.makedirs(out_dir, exist_ok=True)
        out_path = os.path.join(out_dir, f"{basename}.obj")

    # Voorlopig: schrijf enkel vertices als punten
    verts: list[tuple[float, float, float]] = []
    if labels:
        r = 2.0
        for i, _ in enumerate(labels):
            angle = (2 * math.pi * i) / len(labels)
            verts.append((r * math.cos(angle), r * math.sin(angle), 0.0))
    else:
        verts = [(0.0, 0.0, 0.0)]

    with open(out_path, "w", encoding="utf-8") as f:
        for v in verts:
            f.write(f"v {v[0]} {v[1]} {v[2]}\n")

    return out_path



import pathlib
from a003558.export_blender import export_octa_cube_obj, export_label_spheres_obj

def test_export_objs(tmp_path: pathlib.Path):
    out = export_octa_cube_obj(out_dir=str(tmp_path), basename="scene")
    assert all(pathlib.Path(p).exists() for p in out.values())
    lab = export_label_spheres_obj(path=str(tmp_path / "labels.obj"))
    assert pathlib.Path(lab).exists()




from __future__ import annotations

def dyadic_horizon_crossed(n: int, L: int) -> bool:
    """
    Phase trigger: 'universe scale' 2^n crosses critical 2^L.
    """
    return n >= L

def prime_power_spike(m: int) -> bool:
    """
    Heuristic tag: True if m is p^k for odd prime p.
    """
    x = m
    p = None
    d = 2
    found = False
    while d * d <= x:
        e = 0
        while x % d == 0:
            x //= d; e += 1
        if e:
            if p is None: p = d
            else: return False
            found = True
        d = 3 if d == 2 else d + 2
    if x > 1:
        if p is None: p = x; found = True
        else: return False
    return found and (p % 2 == 1)




from __future__ import annotations
import numpy as np

def quenneau_spin_to_dyadic_capacity(pattern: np.ndarray) -> int:
    """
    Toy: given a 'spin' pattern (0/1) length N, return dyadic capacity 2^k >= N
    as minimal cost abstraction target.
    """
    N = len(pattern)
    k = int(np.ceil(np.log2(max(1, N))))
    capacity = 2**k
    return capacity

def compression_loss(pattern: np.ndarray) -> float:
    """
    Simple proxy for loss when coarse-graining to nearest dyadic block.
    """
    N = len(pattern)
    capacity = quenneau_spin_to_dyadic_capacity(pattern)
    return (capacity - N) / capacity




import math
from a003558.number_theory import multiplicative_order
from a003558.horizons import horizon_bound, staircase_level

def test_horizon_bound_holds():
    for n in range(1,1024):
        m = 2*n - 1
        if m == 1: 
            L = 1
        else:
            L = multiplicative_order(2, m)
        assert horizon_bound(n, L)

def test_staircase_level():
    assert staircase_level(1)==1+0
    assert staircase_level(2)==2
    assert staircase_level(3)==2
    assert staircase_level(4)==3




import math
from a003558.number_theory import a003558_order, back_front_cycle_length, multiplicative_order, phi

def test_a003558_small():
    known = [1,2,4,3,6,10,12,4,8]
    for n, val in enumerate(known):
        assert a003558_order(n) == val

def test_back_front_relation():
    for n in range(1,200):
        m = 2*n - 1
        if m == 1:
            assert back_front_cycle_length(n) == 1
        else:
            assert back_front_cycle_length(n) == multiplicative_order(2, m)

def test_order_divides_phi():
    for m in [3,5,7,9,11,13,15,21,25,27,33,35,63,81]:
        if math.gcd(2,m)==1:
            k = multiplicative_order(2, m)
            assert phi(m) % k == 0




from __future__ import annotations
import math
from typing import List, Tuple

def phi(n: int) -> int:
    """Euler's totient via factorization (simple)."""
    if n <= 0: raise ValueError("n must be positive")
    result = n
    m = n
    p = 2
    while p * p <= m:
        if m % p == 0:
            while m % p == 0:
                m //= p
            result -= result // p
        p = 3 if p == 2 else p + 2  # skip evens
    if m > 1:
        result -= result // m
    return result

def multiplicative_order(a: int, m: int) -> int:
    """Smallest k>0 such that a^k ≡ 1 (mod m); assumes gcd(a,m)=1."""
    if math.gcd(a, m) != 1:
        raise ValueError("a and m must be coprime")
    # order divides phi(m): test divisors in increasing order
    t = phi(m)
    # factor t
    fac = {}
    x = t
    d = 2
    while d * d <= x:
        while x % d == 0:
            fac[d] = fac.get(d, 0) + 1
            x //= d
        d = 3 if d == 2 else d + 2
    if x > 1:
        fac[x] = fac.get(x, 0) + 1
    # generate divisors in increasing order
    divs = [1]
    for p, e in fac.items():
        divs = [d * (p ** k) for d in divs for k in range(e + 1)]
    for d in sorted(divs):
        if pow(a, d, m) == 1:
            return d
    raise RuntimeError("order not found")

def a003558_order(n: int) -> int:
    """A003558(n) = ord_{2n+1}(2)."""
    if n < 0:
        raise ValueError("n >= 0 required")
    m = 2*n + 1
    if m == 1:
        return 1
    return multiplicative_order(2, m)

def back_front_cycle_length(n: int) -> int:
    """
    Cycle length for back-front permutation on n items equals ord_{2n-1}(2).
    """
    if n <= 0:
        raise ValueError("n must be >=1")
    m = 2*n - 1
    if m == 1:
        return 1
    return multiplicative_order(2, m)

def prime_power_factors(n: int):
    """Return prime-power factorization list [(p,e),...]."""
    out = []
    m = n
    p = 2
    while p * p <= m:
        if m % p == 0:
            e = 0
            while m % p == 0:
                m //= p
                e += 1
            out.append((p,e))
        p = 3 if p == 2 else p + 2
    if m > 1:
        out.append((m,1))
    return out



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("a003558")
except PackageNotFoundError:
    __version__ = "0"

# Viz niet auto-importeren (vereist matplotlib):
# from a003558.viz import plot_basis  # gebruikers doen dit expliciet



import pathlib
from a003558.viz import plot_basis, plot_cycle
from a003558.viz_octa import plot_octahedron

def test_plot_cycle_smoke(tmp_path: pathlib.Path):
    """
    Smoke-test: plot_cycle moet zonder exception werken en een figuur opslaan.
    """
    out_png = tmp_path / "cycle.png"
    fig = plot_cycle(8, save_path=str(out_png), show=False, title="Cycle Smoke")
    assert out_png.exists()
    assert fig is not None

def test_plot_octahedron_smoke(tmp_path: pathlib.Path):
    """
    Smoke-test: plot_octahedron moet zonder exception werken en een figuur opslaan.
    """
    out_png = tmp_path / "octa.png"
    fig = plot_octahedron(save_path=str(out_png), show=False)
    assert out_png.exists()
    assert fig is not None

def test_plot_basis_again(tmp_path: pathlib.Path):
    """
    Extra check van plot_basis (variant).
    """
    out_png = tmp_path / "basis2.png"
    fig = plot_basis(16, save_path=str(out_png), show=False, title="Basis Again")
    assert out_png.exists()
    assert fig is not None
